/* Dual Port Memory Driver */

#include <asm/byteorder.h>
#include <asm/current.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/uaccess.h>

#include <linux/device.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>

#include <xilinx/xparameters.h>
#if 0
#define BASEADDR XPAR_RAMSTORE2PLB_BASEADDR
#define HOST_INTR XPAR_OPB_INTC_0_SYSTEM2_IRPT_HOST_INTR
#else
#define BASEADDR XPAR_XPS_REGISTERINT_BRAM_BASEADDR
#define HOST_INTR XPAR_XPS_INTC_0_RR_PPC_LINUX_IRPT_HOST_INTR
#endif

#define MODULE_NAME "mossdpm"
MODULE_AUTHOR("GVG");
MODULE_DESCRIPTION("ImageStore DualPortMemory Handler");
MODULE_VERSION("0:0.1-experimental");

static int major = ~0;
static unsigned long condition = 0L; /* 32 condition bits */
static void *wADDR = NULL;
static void *rADDR = NULL;

DECLARE_WAIT_QUEUE_HEAD(waitfld);

static int dpm_init(void);
static void dpm_cleanup(void);
module_init(dpm_init);
module_exit(dpm_cleanup);

/* called when the device file is opened in userspace */
static int dpm_open (struct inode *inode, struct file *file) {
	unsigned int minor = iminor(inode);
	printk(KERN_DEBUG MODULE_NAME ": open minor=%d\n", minor);
	return 0;
}

/* called when the device file is closed in userspace */ 
static int dpm_release (struct inode *inode, struct file *file) {
	unsigned int minor = iminor(inode);
	printk(KERN_DEBUG MODULE_NAME ": release minor=%d\n", minor);
	return 0;
}

/* called when device file is read */
ssize_t dpm_read (struct file *filep, char *buff, size_t count, loff_t *ppos) {
	int ret;
	struct {
	unsigned long host;
	unsigned long engine;
	char ch1[0x200];
	char ch2[0x200];
	char ch3[0x200];
	char ch4[0x200];
	char ch5[0x200];
	char ch6[0x200];
	char ch7[0x200];
	char ch8[0x200-0x8];
	} data;
	char *ptr = NULL;
        unsigned int minor = iminor(filep->f_dentry->d_inode);
        printk(KERN_DEBUG MODULE_NAME ": read, minor=%u count=%u condition=%lu pos=%lu\n", 
				minor, count, condition, (unsigned long)(*ppos));

        /* clear condition bit for this minor number */
        clear_bit(minor, &condition);  /* range is bits 0-31 */

        /* will be awakened by interrupt_handler */
        ret = wait_event_interruptible(waitfld, (test_bit(minor, &condition)));
        if(ret) {
	       	printk(KERN_ALERT MODULE_NAME ": exception -> restart the read\n" );
                return -ERESTARTSYS;
        }

	memset(&data, 0, sizeof(data));
	memcpy_fromio(&data, rADDR, sizeof(data));

	switch(minor) {
	case 0:
		ptr = data.ch1;
        	if(count > strlen(data.ch1)) count = strlen(data.ch1);
        	if(count > sizeof(data)) count = sizeof(data);
        	ret = copy_to_user(buff,ptr,count);
		break;
	case 1:
		ptr = data.ch1;
        	if(count > strlen(data.ch1)) count = strlen(data.ch1);
        	if(count > sizeof(data.ch1)) count = sizeof(data.ch1);
        	ret = copy_to_user(buff,ptr,count);
		break;
	case 2:
		ptr = data.ch2;
        	if(count > strlen(data.ch2)) count = strlen(data.ch2);
        	if(count > sizeof(data.ch2)) count = sizeof(data.ch2);
        	ret = copy_to_user(buff,ptr,count);
		break;
	case 3:
		ptr = data.ch3;
        	if(count > strlen(data.ch3)) count = strlen(data.ch3);
        	if(count > sizeof(data.ch3)) count = sizeof(data.ch3);
        	ret = copy_to_user(buff,ptr,count);
		break;
	case 4:
		ptr = data.ch4;
        	if(count > strlen(data.ch4)) count = strlen(data.ch4);
        	if(count > sizeof(data.ch4)) count = sizeof(data.ch4);
        	ret = copy_to_user(buff,ptr,count);
		break;
	case 5:
		ptr = data.ch5;
        	if(count > strlen(data.ch5)) count = strlen(data.ch5);
        	if(count > sizeof(data.ch5)) count = sizeof(data.ch5);
        	ret = copy_to_user(buff,ptr,count);
		break;
	case 6:
		ptr = data.ch6;
        	if(count > strlen(data.ch6)) count = strlen(data.ch6);
        	if(count > sizeof(data.ch6)) count = sizeof(data.ch6);
        	ret = copy_to_user(buff,ptr,count);
		break;
	case 7:
		ptr = data.ch7;
        	if(count > strlen(data.ch7)) count = strlen(data.ch7);
        	if(count > sizeof(data.ch7)) count = sizeof(data.ch7);
        	ret = copy_to_user(buff,ptr,count);
		break;
	case 8:
		ptr = data.ch8;
        	if(count > strlen(data.ch8)) count = strlen(data.ch8);
        	if(count > sizeof(data.ch8)) count = sizeof(data.ch8);
        	ret = copy_to_user(buff,ptr,count);
		break;
	default:
                return -EFAULT;
	}

        if(ret) {
                printk(KERN_ALERT MODULE_NAME " -> userspace copy failed!\n" );
                return -EFAULT;
        }
        printk(KERN_DEBUG MODULE_NAME " sent %lu bytes\n", (unsigned long)count);

	*ppos += count;
	return count;
}

/** @brief write to register associated with the provided minor number */
ssize_t dpm_write(struct file *filep,const char *buff,size_t count,loff_t *offp)
{
        int err;
	struct {
	unsigned long host;
	unsigned long engine;
	char buf[0x1000-0x8];
	} data;
        unsigned int minor = iminor(filep->f_dentry->d_inode);
        printk(KERN_DEBUG MODULE_NAME ":write minor=%u count=%lu, offset=%lu\n",
                                 minor, (unsigned long)count, (unsigned long)*offp );

	/* limit count to the block size, less the semaphore and byte count words */
        if(count > (sizeof(data.buf)))
                count = (sizeof(data.buf));

	/* clear kernel space buffer, and set the byte count */
	memset(&data,0,sizeof(data));
	data.engine = count;
        set_bit(31, &(data.engine)); /* set interrupt bit */

        /* copy user space buffer to kernel space*/
        err = copy_from_user(data.buf,buff,count);
        if(err) {
                printk(KERN_ALERT MODULE_NAME " -> Kernel copy failed!\n" );
                return -EFAULT;
        }

	/* dump the message to dual port memory */
	memcpy_toio(wADDR, &data, sizeof(data));

        *offp += count;
        return count;
}

/* interrupt handler */
static irqreturn_t dpm_handler(int irq, void *addr, struct pt_regs *regs)
{
	condition = __arch__swab32(ioread32(rADDR));
        clear_bit(31, &condition);
        set_bit(0, &condition);
        barrier();
        wake_up_interruptible_all(&waitfld);
        return IRQ_HANDLED;
}

/* supported file operations are supported */
struct file_operations dpm_fops = {
	.open    = dpm_open,
	.read    = dpm_read,
	.write   = dpm_write,
	.release = dpm_release,
};

/* initialize module (and interrupt) */
static int __init dpm_init(void) {
	int i;
	struct resource *r = NULL;
        condition = 0;

	printk(KERN_DEBUG MODULE_NAME ": initializing\n");
	
	major = register_chrdev (0, MODULE_NAME, &dpm_fops);
	if (!major) return -EIO;
        printk(KERN_INFO MODULE_NAME ": major=%d\n", major);
	
	/* map the "handshake" physical memory region to our space */
	r = request_mem_region( BASEADDR + 0x5000,
					0x1000,
					MODULE_NAME
				);
	if (!r) return -EIO;
	rADDR  = ioremap(BASEADDR + 0x7000, 0x1000);
	if (!rADDR) return -EIO;

	r = request_mem_region( BASEADDR + 0x7000,
					0x1000,
					MODULE_NAME
				);
	if (!r) return -EIO;
	wADDR  = ioremap(BASEADDR + 0x5000, 0x1000);
	if (!wADDR) return -EIO;

        /* INIT AND GENERATE INTERRUPT */
        i = request_irq(HOST_INTR,
                        &dpm_handler, 0, "dpm", rADDR);
        if(i<0) {
                printk(KERN_ALERT "irq %d unavailable\n",
                        HOST_INTR);
                return i;
        }
	return 0;
}

/* release interrupt and cleanup module */
static void __exit dpm_cleanup(void) {
	printk(KERN_DEBUG MODULE_NAME ": terminating\n");
	disable_irq(HOST_INTR);
	free_irq(HOST_INTR, rADDR);
        condition = 0;
	if(rADDR) iounmap(rADDR);
	if(wADDR) iounmap(wADDR);
	release_mem_region(BASEADDR + 0x5000, 0x1000);
	release_mem_region(BASEADDR + 0x7000, 0x1000);
	unregister_chrdev (major, MODULE_NAME);
}

