#ifndef _MOSSIO_H_
#define _MOSSIO_H_

/// \file
/// \copyright 2008 Thomson Grass Valley
/// \author adam clarke
/// \date 2008-06-06
/// \brief a collection of inserters and extractors to support ImageStore operations

typedef struct { int t, l, b, r; } tlbr_t;

typedef struct { int y, u, v; } yuv_t;

/// \brief video channel command serialization functor
/// used to stream a video channel command to an sdi output port device 
class videocmd
{
public:
	typedef enum { f1080i, f720p, f625i, f525i } format;

        videocmd(const long chreg, const size_t wblk, const size_t rblk, const tlbr_t &pos, const tlbr_t &mask, const yuv_t &clr, const yuv_t &bdr, const format fmt) 
			: chreg(chreg), wblk(wblk), rblk(rblk), pos(pos), mask(mask), clr(clr), bdr(bdr), fmt(fmt) {};
        std::ostream &operator()(std::ostream &sink) const; 
private:
        const long chreg; ///< channel register
        const size_t wblk; ///< cache memory block index [range 0 - 2<<25]
        const size_t rblk; ///< cache memory block index [range 0 - 2<<25]
	const tlbr_t &pos; ///< position of the image within the output frame
	const tlbr_t &mask; ///< position of the portion of the output frame that is not masked
	const yuv_t &clr; ///< color of the mask
	const yuv_t &bdr; ///< color of the border
	const format fmt;
};

/// \brief insert a display command
std::ostream &operator<<(std::ostream &out, const videocmd &msg);

#endif //_MOSSIO_H_
