#include <memory.h>

char *bitPack( char *dest, char *src, size_t lines, size_t width, size_t realWidth );
char *bitExpand( char *dest, char *src, size_t lines, size_t width, size_t realWidth );

// Heavily use memmove as it supports moving data where memory overlaps

char *
bitPack( char *dest, char *src, size_t lines, size_t width, size_t realWidth )
{
	size_t	line = 1;										// Start with the first line
	char	*sline = src + realWidth;						// Calculate where the second line starts
	char	*dline = dest + width;							// Calculate where the second line should go

	// Check for in place compression, if differing buffers, move the first line, too (otherwise there is no need to)
	if( dest != src )
		memmove( dest, src, width );
	
	// Starting with the second line, suck them in
	for( ; line < lines; line++, sline += realWidth, dline += width )
		memmove( dline, sline, width );

	return dest;
}

char *
bitExpand( char *dest, char *src, size_t lines, size_t width, size_t realWidth )
{
	size_t	line = lines - 1;								// Start with the last line first
	char	*sline = src + ((lines - 1) * width);			// Calculate where the last line starts
	char	*dline = dest + ((lines - 1) * realWidth);		// Calculate where the last line should go
	
	// Iterate in reverse through all the lines so as to preserve the data
    // sparsely expand the data
	for( ; line ; line--, sline -= width, dline -= realWidth )
		memmove( dline, sline, width );
		
	// Check for in place expansion, if differing buffers, move the first line, too (otherwise there is no need to)
	if( dest != src )
		memmove( dest, src, width );

	return dest;
}

#ifdef DRIVER

#include <stdio.h>

int
main( int ac, char *av[] )
{
    // Cleverly preceed each line with a newline so that the lines print out as lines during tests
	char			s[] =   {	
                                "\n!!!!!!!!!"
                                "\n#########"
                                "\n%%%%%%%%%"
                                "\n*********"
                                "\n@@@@@@@@@" 
                            };

	char			d[sizeof(s)];

    const size_t    lineCount = 5;
	const size_t	paddedSize = sizeof( s ) / lineCount;
    const size_t    realSize = 4;

    // initialize the destination buffer with dots, just so we can see what is actually copied over
	memset( d, '.', sizeof( d ) );
	d[sizeof( d ) - 1] = 0;

	printf( "Initial:\n" );
	printf( "%s\n", s );

	printf( "Compressed:\n" );
	printf( "%s\n", bitPack( d, s, lineCount, realSize, paddedSize ) );
	printf( "\n" );

    // Reverse the initialization (the source is now the destination)
    // but this time, pad it with dots so we can see that it has expanded
	memset( s, '.', sizeof( s ) );
	s[sizeof( s ) - 1] = 0;

	printf( "Expanded:\n" );
	printf( "%s\n", bitExpand( s, d, lineCount, realSize, paddedSize ) );

    return 0;

}

#endif
// vim:ts=4:expandtab:ai
